/*
 * prod_cons LKM 
 * --------------------
 * Producer/Consumer Kernel Module
 */

#include <linux/module.h>  /* Needed by all modules */
#include <linux/kernel.h>  /* Needed for KERN_ALERT */
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/namei.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/mutex.h>
#include <linux/list.h>
#include <linux/version.h>

/* Include the right file according to kernel version  */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#if !defined(DEFINE_SEMAPHORE)
#define DEFINE_SEMAPHORE DECLARE_MUTEX 
#endif
#endif

MODULE_AUTHOR("Me");
MODULE_DESCRIPTION ("producer consumer");
MODULE_LICENSE("GPL");

/* 
 * Name of the /proc entry the module uses to communicate 
 * with user processes
 */
#define PROC_ENTRY_NAME "prod_cons"

/*
 * /proc entry descriptor
 */
static struct proc_dir_entry *prod_cons_entry;

/*
 * List of integers
 */
struct item {
	struct list_head list;
	int integer;
};

LIST_HEAD(integers);

DEFINE_SEMAPHORE(sem);
DEFINE_MUTEX(mymutex);

struct item *build_new_item(int integer) 
{
	struct item *new = kmalloc(sizeof(struct item), GFP_KERNEL);
	if (new) {
		new->integer = integer;
	}
	return new;
}

int get_head(void) {
	int ret;
	struct item *head;
	
	head = list_entry(integers.next, struct item, list);
	ret = head->integer;
	list_del(&head->list);
	kfree(head);
	
	return ret;
}

/*
 * counter_read
 *
 * used to transfer the results of the
 * computations to the user process.
 */
static int 
cons_read(char *buf, char **start, off_t offset, 
				int len, int *eof, void *unused)
{	
	int ret;
	
	if (offset > 0) return 0;

	down (&sem);
	/* This is the consumer's code */
	mutex_lock(&mymutex);
	ret = get_head();
	mutex_unlock(&mymutex);

	*eof = 1;
	return sprintf(buf, "%d\n", ret);
}

/*
 * counter_write
 *
 * Read next digit to add to counter
 */
static int 
prod_write(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	char digit_char;
	int digit;
	struct item *new_item;

	/* Each command can be only a single ASCII digit */
	if (count != 1) {
		return -EFAULT;
	}

	/* copy data from user space */
	if(copy_from_user(&digit_char, buffer, count)) {
		return -EFAULT;
	}
	 
	/* Check if input is a valid digit */
	if (digit_char < '0' || digit_char > '9') {
		return -EFAULT;		
	}
	
	/* convert to integer */
	digit = digit_char - '0';

	/* Produce the new item */
	new_item = build_new_item(digit);
	
	/* Add item in mutual exclusion */
	mutex_lock(&mymutex);
	list_add_tail(&(new_item->list), &integers);
	mutex_unlock(&mymutex);
	
	up (&sem);
	return count;
}

/*
 * init_module - initialize /proc entry
 */
int init_module(void)
{
	prod_cons_entry = create_proc_entry(PROC_ENTRY_NAME, S_IRUGO | S_IWUGO, NULL);
	prod_cons_entry->read_proc = cons_read;
	prod_cons_entry->write_proc = prod_write;
	sema_init(&sem,0);
	return 0;
}


void cleanup_module(void)
{
	if (prod_cons_entry) {
		remove_proc_entry(PROC_ENTRY_NAME, NULL);
	}
}  


